@@ -55,9 +55,14 @@
 #include <sys/user.h>
 #include <sys/eventfd.h>
 #include <time.h>
-#include "cryptouser.h"
+#include <limits.h>
+#include <sys/select.h>
 
+#include <linux/if_alg.h>
+
+#include "cryptouser.h"
 #include "kcapi.h"
+#include "internal.h"
 
 /* remove once in if_alg.h */
 #ifndef ALG_SET_AEAD_ASSOCLEN
@@ -90,6 +95,17 @@
 #define ALG_MAX_PAGES 16
 #endif
 
+#define KCAPI_MAJVERSION 0  /* API / ABI incompatible changes, functional
+			     * changes that require consumer to be updated
+			     * (as long as this number is zero, the API is
+			     * not considered stable and can change without
+			     * a bump of the major version) */
+#define KCAPI_MINVERSION 12 /* API compatible, ABI may change, functional
+			     * enhancements only, consumer can be left
+			     * unchanged if enhancements are not considered */
+#define KCAPI_PATCHLEVEL 0  /* API / ABI compatible, no functional changes, no
+			     * enhancements, bug fixes only */
+
 /************************************************************
  * Declarations for opague data structures
  ************************************************************/
@@ -150,6 +166,39 @@ struct kcapi_aead_data {
 	uint8_t *tag;
 };
 
+/*
+ * This value sets the maximum number of concurrent AIO operations we support.
+ * This value can be changed as needed. However, note that the memory
+ * consumption of one cipher handle increases proportionally to this value. This
+ * means that during the _init API call processing, memory corresponding with
+ * this number must be allocated regardless whether it is used later on or not.
+ *
+ * If the caller supplies more IOVECs to be processed in parallel than this
+ * value, the libkcapi code below segments the the provided input data
+ * into IOVEC chunks of KCAPI_AIO_CONCURRENT size. Thus, the calling user
+ * will not see any difference when this value changes other than the
+ * performance impact during _init (the larger the value, the slower the
+ * _init processing) and later on during the cipher operations (the larger
+ * the value, the more parallel cipher operations are supported).
+ */
+#define KCAPI_AIO_CONCURRENT	64
+
+/**
+ * AIO related data structure to hold all information for AIO
+ * @skcipher_aio_disable: AIO support for symmetric ciphers not present
+ * @efd: event file descriptor
+ * @aio_ctx: AIO context to use for AIO syscalls
+ * @cio: Active concurrent IOCBs
+ */
+struct kcapi_aio {
+	unsigned int skcipher_aio_disable:1;
+	int efd;
+	aio_context_t aio_ctx;
+	uint32_t completed_reads;
+	struct iocb *cio;
+	struct iocb **ciopp;
+};
+
 /**
  * Cipher handle
  * @tfmfd: Socket descriptor for AF_ALG
@@ -177,7 +226,7 @@ struct kcapi_handle {
  ************************************************************/
 static int kcapi_verbosity_level = LOG_ERR;
 
-static void kcapi_dolog(int severity, const char *fmt, ...)
+void kcapi_dolog(int severity, const char *fmt, ...)
 {
 	va_list args;
 	char msg[128];
@@ -212,6 +261,7 @@ static void kcapi_dolog(int severity, const char *fmt, ...)
 	fprintf(stderr, "libkcapi - %s: %s\n", sev, msg);
 }
 
+DSO_PUBLIC
 void kcapi_set_verbosity(enum kcapi_verbosity level)
 {
 	kcapi_verbosity_level = level;
@@ -405,6 +455,10 @@ static inline int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
 	struct iovec iov;
 	uint32_t processed = 0;
 	int ret = 0;
+	uint32_t sflags = (flags & SPLICE_F_MORE) ? MSG_MORE : 0;
+
+	if (inlen > INT_MAX)
+		return -EMSGSIZE;
 
 	ret = _kcapi_common_accept(handle);
 	if (ret)
@@ -417,9 +471,7 @@ static inline int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
 		iov.iov_len = inlen;
 
 		if ((handle->processed_sg++) > ALG_MAX_PAGES) {
-			ret = _kcapi_common_send_data(handle, &iov, 1,
-						      (flags & SPLICE_F_MORE) ?
-						       MSG_MORE : 0);
+			ret = _kcapi_common_send_data(handle, &iov, 1, sflags);
 		} else {
 			ret = vmsplice(handle->pipes[1], &iov, 1,
 				       SPLICE_F_GIFT|flags);
@@ -442,6 +494,152 @@ static inline int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
 	return processed;
 }
 
+/* Wrapper for io_getevents -- returns < 0 on error, or processed bytes */
+static int32_t _kcapi_aio_read_all(struct kcapi_handle *handle, uint32_t toread,
+				   struct timespec *timeout)
+{
+	int32_t processed = 0;
+
+	if (toread > KCAPI_AIO_CONCURRENT)
+		return -EINVAL;
+
+	while (toread) {
+		int i;
+		struct io_event events[KCAPI_AIO_CONCURRENT];
+		int rc = io_getevents(handle->aio.aio_ctx, 1, toread,
+				      events, timeout);
+
+		if (rc < 0)
+			return rc;
+
+		for (i = 0; i < rc; i++) {
+			struct iocb *cb;
+
+			/*
+			 * If one cipher operation fails, so will the entire
+			 * AIO operation
+			 */
+			if (events[i].res < 0)
+				return events[i].res;
+
+			cb = (struct iocb *)(uintptr_t)events[i].obj;
+			processed += cb->aio_nbytes;
+			cb->aio_fildes = 0;
+			handle->aio.completed_reads++;
+		}
+		toread -= rc;
+	}
+
+	return processed;
+}
+
+/* read data from successfully processed cipher operations */
+static int _kcapi_aio_poll_data(struct kcapi_handle *handle, suseconds_t wait)
+{
+	struct timespec timeout;
+	struct timeval tv;
+	fd_set rfds;
+	u_int64_t eval = 0;
+	int ret;
+	int efd = handle->aio.efd;
+
+	FD_ZERO(&rfds);
+	FD_SET(efd, &rfds);
+	tv.tv_sec = 0;
+	tv.tv_usec = wait;
+
+	ret = select(efd + 1, &rfds, NULL, NULL, &tv);
+	if (ret == -1) {
+		kcapi_dolog(LOG_ERR, "Select Error: %d\n", errno);
+		return -errno;
+	}
+	if (!FD_ISSET(efd, &rfds)) {
+		kcapi_dolog(LOG_ERR, "aio poll: no FDs\n");
+		return -EFAULT;
+	}
+
+	if (read(efd, &eval, sizeof(eval)) != sizeof(eval)) {
+		kcapi_dolog(LOG_ERR, "efd read error\n");
+		return -EFAULT;
+	}
+
+	timeout.tv_sec = 0;
+	timeout.tv_nsec = 0;
+
+	return _kcapi_aio_read_all(handle, eval, &timeout);
+}
+
+static int _kcapi_aio_send_iov(struct kcapi_handle *handle,
+			       struct iovec *iov, uint32_t iovlen,
+			       int access, int enc)
+{
+	int ret;
+
+	/*
+	 * Using two syscalls with memcpy is faster than four syscalls
+	 * without memcpy below the given threshold.
+	 */
+	if ((access == KCAPI_ACCESS_HEURISTIC && iov->iov_len <= (1<<13)) ||
+	    access == KCAPI_ACCESS_SENDMSG) {
+		ret = _kcapi_common_send_meta(handle, iov, iovlen, enc, 0);
+		if (0 > ret)
+			return ret;
+	} else {
+		ret = _kcapi_common_send_meta(handle, NULL, 0, enc, MSG_MORE);
+		if (0 > ret)
+			return ret;
+		ret = _kcapi_common_vmsplice_iov(handle, iov, iovlen, 0);
+		if (0 > ret)
+			return ret;
+	}
+
+	return 0;
+}
+
+static int32_t _kcapi_aio_read_iov(struct kcapi_handle *handle,
+				   struct iovec *iov, uint32_t iovlen)
+{
+	struct iocb *cb = handle->aio.cio;
+	uint32_t i;
+	int32_t ret;
+
+	if (iovlen > KCAPI_AIO_CONCURRENT)
+		return -EFAULT;
+
+	for (i = 0; i < iovlen; i++) {
+		while (cb->aio_fildes) {
+			ret = _kcapi_aio_poll_data(handle, 10);
+			if (ret < 0)
+				return ret;
+		}
+
+		memset(cb, 0, sizeof(*cb));
+		cb->aio_fildes = handle->opfd;
+		cb->aio_lio_opcode = IOCB_CMD_PREAD;
+		cb->aio_buf = (unsigned long)iov->iov_base;
+		cb->aio_offset = 0;
+		cb->aio_data = i;
+		cb->aio_nbytes = iov->iov_len;
+		cb->aio_flags = IOCB_FLAG_RESFD;
+		cb->aio_resfd = handle->aio.efd;
+		cb++;
+		iov++;
+	}
+
+	ret = io_submit(handle->aio.aio_ctx, iovlen, handle->aio.ciopp);
+	if ((uint32_t)ret != iovlen) {
+		if (ret < 0) {
+			kcapi_dolog(LOG_ERR, "io_read Error: %d\n", errno);
+			return -EFAULT;
+		} else {
+			kcapi_dolog(LOG_ERR, "Could not sumbit AIO read\n");
+			return -EIO;
+		}
+	}
+
+	return _kcapi_aio_poll_data(handle, 1);
+}
+
 static inline int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 					      struct iovec *iov,
 					      uint32_t iovlen)
@@ -467,15 +665,6 @@ static inline int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 	kcapi_dolog(LOG_DEBUG, "AF_ALG: recvmsg syscall returned %d (errno: %d)",
 		    ret, errsv);
 
-#if 0
-	/*
-	 * Truncated message digests can be identified with this check.
-	 */
-	if (msg.msg_flags & MSG_TRUNC) {
-		fprintf(stderr, "recvmsg: processed data was truncated by kernel (only %lu bytes processed)\n", (unsigned long)ret);
-		return -EMSGSIZE;
-	}
-#endif
 
 	/*
 	 * As the iovecs are processed and removed from the list in the kernel
@@ -488,6 +677,16 @@ static inline int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 	if (ret >= 0 || errsv == EBADMSG)
 		handle->processed_sg = 0;
 
+#if 0
+	/*
+	 * Truncated message digests can be identified with this check.
+	 */
+	if (msg.msg_flags & MSG_TRUNC) {
+		fprintf(stderr, "recvmsg: processed data was truncated by kernel (only %lu bytes processed)\n", (unsigned long)ret);
+		return -EMSGSIZE;
+	}
+#endif
+
 	return (ret >= 0) ? ret : -errsv;
 }
 
@@ -497,6 +696,9 @@ static inline int32_t _kcapi_common_read_data(struct kcapi_handle *handle,
 	int32_t ret = 0;
 	int32_t errsv;
 
+	if (outlen > INT_MAX)
+		return -EMSGSIZE;
+
 	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
@@ -742,33 +944,6 @@ static int _kcapi_common_getinfo(struct kcapi_handle *handle,
 	return 0;
 }
 
-/* Initialize AIO -- on error, AIO is simply not used, but lib lives on */
-static void _kcapi_aio_init(struct kcapi_handle *handle)
-{
-	if (handle->aio.skcipher_aio_disable)
-		return;
-
-	handle->aio.efd = eventfd(0, EFD_CLOEXEC);
-	if (handle->aio.efd < 0)
-		goto err;
-
-	if (io_setup(KCAPI_AIO_CONCURRENT, &handle->aio.aio_ctx) < 0) {
-		kcapi_dolog(LOG_ERR, "io_setup error");
-		goto err;
-	}
-
-	kcapi_dolog(LOG_VERBOSE, "asynchronoous I/O initialized");
-
-	return;
-
-err:
-	handle->aio.skcipher_aio_disable = 1;
-	if (handle->aio.efd != -1)
-		close(handle->aio.efd);
-	handle->aio.efd = -1;
-	return;
-}
-
 static inline void _kcapi_aio_destroy(struct kcapi_handle *handle)
 {
 	if (handle->aio.skcipher_aio_disable)
@@ -776,10 +951,18 @@ static inline void _kcapi_aio_destroy(struct kcapi_handle *handle)
 	if (handle->aio.efd != -1)
 		close(handle->aio.efd);
 	io_destroy(handle->aio.aio_ctx);
+	if (handle->aio.cio)
+		free(handle->aio.cio);
+	handle->aio.cio = NULL;
+	if (handle->aio.ciopp)
+		free(handle->aio.ciopp);
+	handle->aio.ciopp = NULL;
 }
 
 static inline void _kcapi_handle_destroy(struct kcapi_handle *handle)
 {
+	if (!handle)
+		return;
 	if (handle->tfmfd != -1)
 		close(handle->tfmfd);
 	if (handle->opfd != -1)
@@ -792,6 +975,62 @@ static inline void _kcapi_handle_destroy(struct kcapi_handle *handle)
 	memset(handle, 0, sizeof(struct kcapi_handle));
 }
 
+static int _kcapi_aio_init(struct kcapi_handle *handle)
+{
+	uint32_t i;
+	int err;
+
+	handle->aio.cio = calloc(KCAPI_AIO_CONCURRENT, sizeof(struct iocb));
+	if (!handle->aio.cio)
+		return ENOMEM;
+
+	handle->aio.ciopp = calloc(KCAPI_AIO_CONCURRENT, sizeof(void *));
+	if (!handle->aio.ciopp) {
+		err = ENOMEM;
+		goto err;
+	}
+
+	/*
+	 * Set up the pointers to pointers array that is required by
+	 * io_submit. Please do not ask me why the kernel wants this. :-)
+	 */
+	for (i = 0; i < KCAPI_AIO_CONCURRENT; i++)
+		*(handle->aio.ciopp + i) = handle->aio.cio + i;
+
+	handle->aio.efd = eventfd(0, EFD_CLOEXEC);
+	if (handle->aio.efd < 0) {
+		err = errno;
+		kcapi_dolog(LOG_ERR, "Event FD cannot be initialized: %d\n",
+			    err);
+		goto err;
+	}
+
+	err = io_setup(KCAPI_AIO_CONCURRENT, &handle->aio.aio_ctx);
+	if (err < 0) {
+		kcapi_dolog(LOG_ERR, "io_setup error %d\n", err);
+		/* turn return code into an errno */
+		err = -err;
+		goto err;
+	}
+
+	kcapi_dolog(LOG_VERBOSE, "asynchronoous I/O initialized");
+
+	return 0;
+
+err:
+	handle->aio.skcipher_aio_disable = 1;
+	if (handle->aio.efd != -1)
+		close(handle->aio.efd);
+	handle->aio.efd = -1;
+	if (handle->aio.cio)
+		free(handle->aio.cio);
+	handle->aio.cio = NULL;
+	if (handle->aio.ciopp)
+		free(handle->aio.ciopp);
+	handle->aio.ciopp = NULL;
+	return err;
+}
+
 static int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 			      const char *ciphername, uint32_t flags)
 {
@@ -809,38 +1048,40 @@ static int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 	if (!handle)
 		return -ENOMEM;
 
+	handle->opfd = -1;
+	handle->tfmfd = -1;
+	handle->pipes[0] = -1;
+	handle->pipes[1] = -1;
+
 	memset(&sa, 0, sizeof(sa));
 	sa.salg_family = AF_ALG;
 	snprintf((char *)sa.salg_type, sizeof(sa.salg_type),"%s", type);
 	snprintf((char *)sa.salg_name, sizeof(sa.salg_name),"%s", ciphername);
 
 	handle->tfmfd = socket(AF_ALG, SOCK_SEQPACKET, 0);
-	kcapi_dolog(LOG_DEBUG, "AF_ALG: socket syscall %s (errno: %d)",
-		    (handle->tfmfd == -1) ? "failed": "passed", errno);
 	if (handle->tfmfd == -1) {
-		errsv = -EOPNOTSUPP;
+		errsv = errno;
+		kcapi_dolog(LOG_ERR, "AF_ALG: socket syscall failed (errno: %d)",
+			    errsv);
 		goto err;
 	}
+	kcapi_dolog(LOG_DEBUG, "AF_ALG: socket syscall passed (errno: %d)",
+		    errno);
 
 	if (bind(handle->tfmfd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
 		errsv = errno;
 		kcapi_dolog(LOG_ERR, "AF_ALG: bind failed (errno: %d)",
 			    errsv);
-		close(handle->tfmfd);
-		handle->tfmfd = -1;
 		goto err;
 	}
 	kcapi_dolog(LOG_DEBUG, "AF_ALG: bind syscall passed (errno: %d)",
 		    errno);
 
-	handle->opfd = -1;
-
 	ret = pipe(handle->pipes);
 	if (ret) {
 		errsv = errno;
-		kcapi_dolog(LOG_DEBUG, "AF_ALG: pipe syscall failed (errno: %d)",
+		kcapi_dolog(LOG_ERR, "AF_ALG: pipe syscall failed (errno: %d)",
 			    errsv);
-		close(handle->tfmfd);
 		goto err;
 	}
 	kcapi_dolog(LOG_DEBUG, "AF_ALG: pipe syscall passed");
@@ -853,9 +1094,11 @@ static int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 		goto err;
 	}
 
-	if (flags & KCAPI_INIT_AIO)
-		_kcapi_aio_init(handle);
-	else
+	if (flags & KCAPI_INIT_AIO) {
+		errsv = _kcapi_aio_init(handle);
+		if (errsv)
+			goto err;
+	} else
 		handle->aio.skcipher_aio_disable = 1;
 
 	kcapi_dolog(LOG_VERBOSE, "communication for %s with kernel initialized", ciphername);
@@ -871,29 +1114,33 @@ err:
 
 /*********** Generic Helper functions *************************/
 
+DSO_PUBLIC
 void kcapi_memset_secure(void *s, int c, uint32_t n)
 {
 	memset(s, c, n);
 	__asm__ __volatile__("" : : "r" (s) : "memory");
 }
 
+DSO_PUBLIC
 void kcapi_versionstring(char *buf, uint32_t buflen)
 {
-	snprintf(buf, buflen, "libkcapi %d.%d.%d", MAJVERSION, MINVERSION,
-		 PATCHLEVEL);
+	snprintf(buf, buflen, "libkcapi %d.%d.%d", KCAPI_MAJVERSION,
+		 KCAPI_MINVERSION, KCAPI_PATCHLEVEL);
 }
 
+DSO_PUBLIC
 uint32_t kcapi_version(void)
 {
 	uint32_t version = 0;
 
-	version =  MAJVERSION * 1000000;
-	version += MINVERSION * 10000;
-	version += PATCHLEVEL * 100;
+	version =  KCAPI_MAJVERSION * 1000000;
+	version += KCAPI_MINVERSION * 10000;
+	version += KCAPI_PATCHLEVEL * 100;
 
 	return version;
 }
 
+DSO_PUBLIC
 int kcapi_pad_iv(struct kcapi_handle *handle,
 		 const uint8_t *iv, uint32_t ivlen,
 		 uint8_t **newiv, uint32_t *newivlen)
@@ -916,12 +1163,14 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 	return 0;
 }
 
+DSO_PUBLIC
 int kcapi_cipher_init(struct kcapi_handle **handle, const char *ciphername,
 		      uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "skcipher", ciphername, flags);
 }
 
+DSO_PUBLIC
 void kcapi_cipher_destroy(struct kcapi_handle *handle)
 {
 	if (!handle)
@@ -930,55 +1179,13 @@ void kcapi_cipher_destroy(struct kcapi_handle *handle)
 	kcapi_memset_secure(handle, 0, sizeof(struct kcapi_handle));
 }
 
+DSO_PUBLIC
 int kcapi_cipher_setkey(struct kcapi_handle *handle,
 			const uint8_t *key, uint32_t keylen)
 {
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
-static inline int32_t _kcapi_aio_read(struct kcapi_handle *handle,
-				     uint8_t *out, uint32_t outlen)
-{
-	(void)handle;
-	(void)out;
-	(void)outlen;
-	kcapi_dolog(LOG_ERR, "AIO support not complete");
-	return -EOPNOTSUPP;
-#if 0
-	struct iocb *cb = NULL;
-	int r = 0;
-
-	cb = &cbt[handle->aio.curr_cio];
-        while (cb->aio_fildes)
-		poll_data(10);
-
-        memset(cb, 0, sizeof(*cb));
-        cb->aio_fildes = handle->opfd;
-        cb->aio_lio_opcode = IOCB_CMD_PREAD;
-        cb->aio_buf = out;
-        cb->aio_offset = 0;
-        cb->aio_data = i;
-        cb->aio_nbytes = outlen;
-        cb->aio_flags = IOCB_FLAG_RESFD;
-        cb->aio_resfd = handle->aio.efd;
-        r = io_read(handle->aio.aio_ctx, 1, &cb);
-        if (r != 1) {
-		if (r < 0) {
-			fprintf(stderr, "io_read Error: %d\n", errno);
-			return -EFAULT;
-		} else {
-			fprintf(stderr, "Could not sumbit AIO read\n");
-			return -EIO;
-		}
-	}
-
-        if (handle->aio.curr_cio) >= KCAPI_AIO_CONCURRENT / 2)
-		poll_data(1);
-
-	return 0;
-#endif
-}
-
 static int32_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
 				   const uint8_t *in, uint32_t inlen,
 				   uint8_t *out, uint32_t outlen,
@@ -987,6 +1194,9 @@ static int32_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
 	struct iovec iov;
 	int32_t ret = 0;
 
+	if (outlen > INT_MAX)
+		return -EMSGSIZE;
+
 	if (!in || !inlen || !out || !outlen) {
 		kcapi_dolog(LOG_ERR,
 			    "Symmetric Encryption: Empty plaintext or ciphertext buffer provided");
@@ -1005,7 +1215,7 @@ static int32_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
 		if (0 > ret)
 			return ret;
 	} else {
-		ret = _kcapi_common_send_meta(handle, NULL, 0, enc, 0);
+		ret = _kcapi_common_send_meta(handle, NULL, 0, enc, MSG_MORE);
 		if (0 > ret)
 			return ret;
 		ret = _kcapi_common_vmsplice_chunk(handle, in, inlen, 0);
@@ -1013,18 +1223,7 @@ static int32_t _kcapi_cipher_crypt(struct kcapi_handle *handle,
 			return ret;
 	}
 
-	if (handle->aio.skcipher_aio_disable) {
-		return _kcapi_common_read_data(handle, out, outlen);
-	} else {
-		ret = _kcapi_aio_read(handle, out, outlen);
-		if (ret == -EIO) {
-			/* AIO not available for specific interface */
-			handle->aio.skcipher_aio_disable = 1;
-			_kcapi_aio_destroy(handle);
-			return _kcapi_common_read_data(handle, out, outlen);
-		}
-		return ret;
-	}
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 
 static int32_t _kcapi_cipher_crypt_chunk(struct kcapi_handle *handle,
@@ -1033,9 +1232,14 @@ static int32_t _kcapi_cipher_crypt_chunk(struct kcapi_handle *handle,
 					int access, int enc)
 {
 	int32_t totallen = 0;
+	uint32_t maxprocess = sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES;
+
+	if (outlen > INT_MAX)
+		return -EMSGSIZE;
 
 	while (inlen) {
-		uint32_t process = inlen;
+		uint32_t inprocess = inlen;
+		uint32_t outprocess = outlen;
 		int32_t ret = 0;
 
 		/*
@@ -1043,17 +1247,19 @@ static int32_t _kcapi_cipher_crypt_chunk(struct kcapi_handle *handle,
 		 * a multiple of blocksize, because we assume that this is
 		 * always the case.
 		 */
-		if (inlen > (uint32_t)(sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES))
-			process = sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES;
+		if (inlen > maxprocess)
+			inprocess = maxprocess;
+		if (outlen > maxprocess)
+			outprocess = maxprocess;
 
-		ret = _kcapi_cipher_crypt(handle, in, process, out, outlen,
-					  access, enc);
+		ret = _kcapi_cipher_crypt(handle, in, inprocess, out,
+					  outprocess, access, enc);
 		if (ret < 0)
 			return ret;
 
-		totallen += process;
-		in += process;
-		inlen -= process;
+		totallen += inprocess;
+		in += inprocess;
+		inlen -= inprocess;
 		out += ret;
 		outlen -= ret;
 	}
@@ -1061,6 +1267,67 @@ static int32_t _kcapi_cipher_crypt_chunk(struct kcapi_handle *handle,
 	return totallen;
 }
 
+static int32_t _kcapi_cipher_crypt_aio(struct kcapi_handle *handle,
+				       struct iovec *iov, uint32_t iovlen,
+				       int access, int enc)
+{
+	int32_t ret;
+	int32_t rc;
+	uint32_t tosend = iovlen;
+
+	if (handle->aio.skcipher_aio_disable) {
+		kcapi_dolog(LOG_WARN, "AIO support disabled\n");
+		return -EOPNOTSUPP;
+	}
+
+	ret = _kcapi_common_accept(handle);
+	if (ret)
+		return ret;
+
+	handle->aio.completed_reads = 0;
+
+	/* Every IOVEC is processed as its individual cipher operation. */
+	while (tosend) {
+		uint32_t process = KCAPI_AIO_CONCURRENT < tosend ?
+					KCAPI_AIO_CONCURRENT : tosend;
+		int32_t rc = _kcapi_aio_send_iov(handle, iov, process,
+						 access, enc);
+
+		if (rc < 0)
+			return rc;
+
+		rc = _kcapi_aio_read_iov(handle, iov, process);
+		if (rc < 0)
+			return rc;
+
+		iov += process;
+		ret += rc;
+		tosend -= handle->aio.completed_reads;
+	}
+
+	/*
+	 * If a multi-staged AIO operation shall be designed, the following
+	 * loop needs to be moved to a closing API call. If done so, the
+	 * current function could be invoked multiple times to send more data
+	 * to the kernel before the closing call requires that all outstanding
+	 * requests are to be completed.
+	 *
+	 * If a multi-staged AIO operation is to be implemented, the issue
+	 * is that when submitting a number of requests, the caller is not
+	 * able to detect which particular request is completed. Thus, an
+	 * "open-ended" multi-staged AIO operation could not be implemented.
+	 */
+
+	rc = _kcapi_aio_read_all(handle, iovlen - handle->aio.completed_reads,
+				 NULL);
+	if (rc < 0)
+		return rc;
+	ret += rc;
+
+	return ret;
+}
+
+DSO_PUBLIC
 int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
 			     const uint8_t *iv,
@@ -1081,6 +1348,16 @@ int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 					 ALG_OP_ENCRYPT);
 }
 
+DSO_PUBLIC
+int32_t kcapi_cipher_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+				 uint32_t iovlen, const uint8_t *iv, int access)
+{
+	handle->cipher.iv = iv;
+	return _kcapi_cipher_crypt_aio(handle, iov, iovlen, access,
+				       ALG_OP_ENCRYPT);
+}
+
+DSO_PUBLIC
 int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
 			     const uint8_t *iv,
@@ -1106,6 +1383,16 @@ int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 					 ALG_OP_DECRYPT);
 }
 
+DSO_PUBLIC
+int32_t kcapi_cipher_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+				 uint32_t iovlen, const uint8_t *iv, int access)
+{
+	handle->cipher.iv = iv;
+	return _kcapi_cipher_crypt_aio(handle, iov, iovlen, access,
+				       ALG_OP_DECRYPT);
+}
+
+DSO_PUBLIC
 int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 				     const uint8_t *iv,
 				     struct iovec *iov, uint32_t iovlen)
@@ -1115,6 +1402,7 @@ int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 				     const uint8_t *iv,
 				     struct iovec *iov, uint32_t iovlen)
@@ -1124,6 +1412,7 @@ int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
 				   struct iovec *iov, uint32_t iovlen)
 {
@@ -1134,6 +1423,7 @@ int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
 		return _kcapi_common_send_data(handle, iov, iovlen, MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
 			       struct iovec *iov, uint32_t iovlen)
 {
@@ -1145,33 +1435,39 @@ int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
 	return _kcapi_common_recv_data(handle, iov, iovlen);
 }
 
+DSO_PUBLIC
 uint32_t kcapi_cipher_ivsize(struct kcapi_handle *handle)
 {
 	return handle->info.ivsize;
 }
 
+DSO_PUBLIC
 uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle)
 {
 	return handle->info.blocksize;
 }
 
+DSO_PUBLIC
 int kcapi_aead_init(struct kcapi_handle **handle, const char *ciphername,
 		    uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "aead", ciphername, flags);
 }
 
+DSO_PUBLIC
 void kcapi_aead_destroy(struct kcapi_handle *handle)
 {
 	_kcapi_handle_destroy(handle);
 }
 
+DSO_PUBLIC
 int kcapi_aead_setkey(struct kcapi_handle *handle,
 		      const uint8_t *key, uint32_t keylen)
 {
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
+DSO_PUBLIC
 int kcapi_aead_settaglen(struct kcapi_handle *handle, uint32_t taglen)
 {
 	handle->aead.tag = NULL;
@@ -1183,11 +1479,13 @@ int kcapi_aead_settaglen(struct kcapi_handle *handle, uint32_t taglen)
 	return 0;
 }
 
+DSO_PUBLIC
 void kcapi_aead_setassoclen(struct kcapi_handle *handle, uint32_t assoclen)
 {
 	handle->aead.assoclen = assoclen;
 }
 
+DSO_PUBLIC
 void kcapi_aead_getdata(struct kcapi_handle *handle,
 			uint8_t *encdata, uint32_t encdatalen,
 			uint8_t **aad, uint32_t *aadlen,
@@ -1216,12 +1514,12 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
 	}
 }
 
+DSO_PUBLIC
 int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
 			   const uint8_t *iv,
 			   uint8_t *out, uint32_t outlen, int access)
 {
-	struct iovec iov;
 	int32_t ret = 0;
 
 	/* require properly sized output data size */
@@ -1240,26 +1538,8 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	}
 
 	handle->cipher.iv = iv;
-
-	iov.iov_base = in;
-	iov.iov_len = inlen;
-	if (access == KCAPI_ACCESS_HEURISTIC ||
-	    access == KCAPI_ACCESS_SENDMSG) {
-		ret = _kcapi_common_send_meta(handle, &iov, 1,
-					      ALG_OP_ENCRYPT, 0);
-		if (0 > ret)
-			return ret;
-	} else {
-		ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
-			inlen ? MSG_MORE : 0);
-		if (0 > ret)
-			return ret;
-		ret = _kcapi_common_vmsplice_iov(handle, &iov, 1, 0);
-		if (0 > ret)
-			return ret;
-	}
-
-	ret = _kcapi_common_read_data(handle, out, outlen);
+	ret = _kcapi_cipher_crypt(handle, in, inlen, out, outlen, access,
+				  ALG_OP_ENCRYPT);
 	if (ret < 0)
 		return ret;
 	if ((ret < (int32_t)handle->aead.taglen))
@@ -1268,22 +1548,41 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	return ret;
 }
 
+DSO_PUBLIC
+int32_t kcapi_aead_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+			       uint32_t iovlen, const uint8_t *iv, int access)
+{
+	int32_t ret = 0;
+	uint32_t i;
+
+	handle->cipher.iv = iv;
+
+	/*
+	 * Currently the kernel is only able to handle one complete individual
+	 * AEAD cipher operation at a time.
+	 *
+	 * The key to this limitation lies in the check (usedpages < outlen)
+	 * in the function aead_recvmsg_async.
+	 */
+	for (i = 0; i < iovlen; i++) {
+		int32_t rc = _kcapi_cipher_crypt_aio(handle, &iov[i], 1,
+						     access, ALG_OP_ENCRYPT);
+
+		if (rc < 0)
+			return rc;
+		ret += rc;
+	}
+
+	return ret;
+}
+
+DSO_PUBLIC
 int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
 			   const uint8_t *iv,
 			   uint8_t *out, uint32_t outlen, int access)
 {
-	struct iovec iov;
-	int32_t ret = 0;
-
 	/* require properly sized output data size */
-	if (outlen < inlen) {
-		kcapi_dolog(LOG_ERR,
-			    "AEAD Decryption: Plaintext buffer (%u) is smaller than ciphertext buffer (%u)",
-			    outlen, inlen);
-		return -EINVAL;
-	}
-
 	if (inlen > (uint32_t)(sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES)) {
 		kcapi_dolog(LOG_ERR,
 			    "AEAD Decryption: Ciphertext buffer (%u) is larger than maximum chunk size (%lu)",
@@ -1292,35 +1591,39 @@ int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 	}
 
 	handle->cipher.iv = iv;
-
-	iov.iov_base = in;
-	iov.iov_len = inlen;
-	if (access == KCAPI_ACCESS_HEURISTIC ||
-	    access == KCAPI_ACCESS_SENDMSG) {
-		ret = _kcapi_common_send_meta(handle, &iov, 1,
-					      ALG_OP_DECRYPT, 0);
-		if (0 > ret)
-			return ret;
-	} else {
-		ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT,
-					      MSG_MORE);
-		if (0 > ret)
-			return ret;
-		ret = _kcapi_common_vmsplice_iov(handle, &iov, 1, 0);
-		if (0 > ret)
-			return ret;
-	}
-
-	/* handle null test vectors */
-	if (outlen) {
-		return _kcapi_common_read_data(handle, out, outlen);
-	} else {
-		iov.iov_base = out;
-		iov.iov_len = outlen;
-		return _kcapi_common_recv_data(handle, &iov, 1);
-	}
+	return _kcapi_cipher_crypt(handle, in, inlen, out, outlen, access,
+				  ALG_OP_DECRYPT);
 }
 
+DSO_PUBLIC
+int32_t kcapi_aead_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+			       uint32_t iovlen, const uint8_t *iv, int access)
+{
+	int32_t ret = 0;
+	uint32_t i;
+
+	handle->cipher.iv = iv;
+
+	/*
+	 * Currently the kernel is only able to handle one complete individual
+	 * AEAD cipher operation at a time.
+	 *
+	 * The key to this limitation lies in the check (usedpages < outlen)
+	 * in the function aead_recvmsg_async.
+	 */
+	for (i = 0; i < iovlen; i++) {
+		int32_t rc = _kcapi_cipher_crypt_aio(handle, &iov[i], 1,
+						     access, ALG_OP_DECRYPT);
+
+		if (rc < 0)
+			return rc;
+		ret += rc;
+	}
+
+	return ret;
+}
+
+DSO_PUBLIC
 int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 				   const uint8_t *iv,
 				   struct iovec *iov, uint32_t iovlen)
@@ -1330,6 +1633,7 @@ int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 				   const uint8_t *iv,
 				   struct iovec *iov, uint32_t iovlen)
@@ -1339,6 +1643,7 @@ int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
 				 struct iovec *iov, uint32_t iovlen)
 {
@@ -1349,6 +1654,7 @@ int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
 		return _kcapi_common_send_data(handle, iov, iovlen, MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
 				      struct iovec *iov, uint32_t iovlen)
 {
@@ -1358,6 +1664,7 @@ int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
 		return _kcapi_common_send_data(handle, iov, iovlen, 0);
 }
 
+DSO_PUBLIC
 int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			     struct iovec *iov, uint32_t iovlen)
 {
@@ -1370,21 +1677,25 @@ int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 	return _kcapi_common_recv_data(handle, iov, iovlen);
 }
 
+DSO_PUBLIC
 uint32_t kcapi_aead_ivsize(struct kcapi_handle *handle)
 {
 	return handle->info.ivsize;
 }
 
+DSO_PUBLIC
 uint32_t kcapi_aead_blocksize(struct kcapi_handle *handle)
 {
 	return handle->info.blocksize;
 }
 
+DSO_PUBLIC
 uint32_t kcapi_aead_authsize(struct kcapi_handle *handle)
 {
 	return handle->info.aead_maxauthsize;
 }
 
+DSO_PUBLIC
 uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
 			    uint32_t inlen, uint32_t assoclen, uint32_t taglen)
 {
@@ -1393,6 +1704,7 @@ uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
 	return ((inlen + bs - 1) / bs * bs + taglen + assoclen);
 }
 
+DSO_PUBLIC
 int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
 			       uint8_t **iv, uint32_t *ivlen)
 {
@@ -1415,17 +1727,20 @@ int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
 	return 0;
 }
 
+DSO_PUBLIC
 int kcapi_md_init(struct kcapi_handle **handle, const char *ciphername,
 		  uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "hash", ciphername, flags);
 }
 
+DSO_PUBLIC
 void kcapi_md_destroy(struct kcapi_handle *handle)
 {
 	_kcapi_handle_destroy(handle);
 }
 
+DSO_PUBLIC
 int kcapi_md_setkey(struct kcapi_handle *handle,
 		    const uint8_t *key, uint32_t keylen)
 {
@@ -1437,6 +1752,9 @@ static inline int32_t _kcapi_md_update(struct kcapi_handle *handle,
 {
 	int32_t ret = 0;
 
+	if (len > INT_MAX)
+		return -EMSGSIZE;
+
 	/* zero buffer length cannot be handled via splice */
 	if (len < (1<<15)) {
 		ret = _kcapi_common_accept(handle);
@@ -1453,6 +1771,7 @@ static inline int32_t _kcapi_md_update(struct kcapi_handle *handle,
 	return 0;
 }
 
+DSO_PUBLIC
 int32_t kcapi_md_update(struct kcapi_handle *handle,
 			const uint8_t *buffer, uint32_t len)
 {
@@ -1476,12 +1795,14 @@ static int32_t _kcapi_md_final(struct kcapi_handle *handle,
 	return _kcapi_common_recv_data(handle, &iov, 1);
 }
 
+DSO_PUBLIC
 int32_t kcapi_md_final(struct kcapi_handle *handle,
 		       uint8_t *buffer, uint32_t len)
 {
 	return _kcapi_md_final(handle, buffer, len);
 }
 
+DSO_PUBLIC
 int32_t kcapi_md_digest(struct kcapi_handle *handle,
 		       const uint8_t *in, uint32_t inlen,
 		       uint8_t *out, uint32_t outlen)
@@ -1494,33 +1815,39 @@ int32_t kcapi_md_digest(struct kcapi_handle *handle,
 	return _kcapi_md_final(handle, out, outlen);
 }
 
+DSO_PUBLIC
 uint32_t kcapi_md_digestsize(struct kcapi_handle *handle)
 {
 	return handle->info.hash_digestsize;
 }
 
+DSO_PUBLIC
 uint32_t kcapi_md_blocksize(struct kcapi_handle *handle)
 {
 	return handle->info.blocksize;
 }
 
+DSO_PUBLIC
 int kcapi_rng_init(struct kcapi_handle **handle, const char *ciphername,
 		   uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "rng", ciphername, flags);
 }
 
+DSO_PUBLIC
 void kcapi_rng_destroy(struct kcapi_handle *handle)
 {
 	_kcapi_handle_destroy(handle);
 }
 
+DSO_PUBLIC
 int kcapi_rng_seed(struct kcapi_handle *handle, uint8_t *seed,
 		   uint32_t seedlen)
 {
 	return _kcapi_common_setkey(handle, seed, seedlen);
 }
 
+DSO_PUBLIC
 int32_t kcapi_rng_generate(struct kcapi_handle *handle,
 			   uint8_t *buffer, uint32_t len)
 {
@@ -1544,23 +1871,27 @@ int32_t kcapi_rng_generate(struct kcapi_handle *handle,
 	return out;
 }
 
+DSO_PUBLIC
 int kcapi_akcipher_init(struct kcapi_handle **handle, const char *ciphername,
 			uint32_t flags)
 {
 	return _kcapi_handle_init(handle, "akcipher", ciphername, flags);
 }
 
+DSO_PUBLIC
 void kcapi_akcipher_destroy(struct kcapi_handle *handle)
 {
 	_kcapi_handle_destroy(handle);
 }
 
+DSO_PUBLIC
 int kcapi_akcipher_setkey(struct kcapi_handle *handle,
 			  const uint8_t *key, uint32_t keylen)
 {
 	return _kcapi_common_setkey(handle, key, keylen);
 }
 
+DSO_PUBLIC
 int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
 			     const uint8_t *key, uint32_t keylen)
 {
@@ -1570,6 +1901,7 @@ int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
 	return (ret >= 0) ? ret : -errno;
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
 			       uint8_t *out, uint32_t outlen, int access)
@@ -1578,6 +1910,7 @@ int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
 				   ALG_OP_ENCRYPT);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
 			       uint8_t *out, uint32_t outlen, int access)
@@ -1586,6 +1919,7 @@ int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
 				   ALG_OP_DECRYPT);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
 			    const uint8_t *in, uint32_t inlen,
 			    uint8_t *out, uint32_t outlen, int access)
@@ -1594,6 +1928,7 @@ int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
 				   ALG_OP_SIGN);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
 			      const uint8_t *in, uint32_t inlen,
 			      uint8_t *out, uint32_t outlen, int access)
@@ -1602,6 +1937,7 @@ int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
 				   ALG_OP_VERIFY);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen)
 {
@@ -1609,6 +1945,7 @@ int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen)
 {
@@ -1616,6 +1953,7 @@ int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen)
 {
@@ -1623,6 +1961,7 @@ int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen)
 {
@@ -1630,6 +1969,7 @@ int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
 				       MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
 				     struct iovec *iov, uint32_t iovlen)
 {
@@ -1641,6 +1981,7 @@ int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
 		return _kcapi_common_send_data(handle, iov, iovlen, MSG_MORE);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
 				          struct iovec *iov, uint32_t iovlen)
 {
@@ -1651,6 +1992,7 @@ int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
 		return _kcapi_common_send_data(handle, iov, iovlen, 0);
 }
 
+DSO_PUBLIC
 int32_t kcapi_akcipher_stream_op(struct kcapi_handle *handle,
 			         struct iovec *iov, uint32_t iovlen)
 {
