@@ -38,25 +38,11 @@
 #define _KCAPI_H
 
 #include <stdint.h>
-
-#include <linux/if_alg.h>
 #include <sys/uio.h>
-#include "kcapi_aio.h"
 
-#define MAJVERSION 0  /* API / ABI incompatible changes, functional changes that
-		       * require consumer to be updated (as long as this number
-		       * is zero, the API is not considered stable and can
-		       * change without a bump of the major version) */
-#define MINVERSION 11 /* API compatible, ABI may change, functional
-		       * enhancements only, consumer can be left unchanged if
-		       * enhancements are not considered */
-#define PATCHLEVEL 1  /* API / ABI compatible, no functional changes, no
-		       * enhancements, bug fixes only */
-
-#if __GNUC__ >= 4
-# define DSO_PUBLIC __attribute__ ((visibility ("default")))
-#else
-# define DSO_PUBLIC
+#ifdef __cplusplus
+extern "C"
+{
 #endif
 
 /*
@@ -107,7 +93,6 @@ struct kcapi_handle;
  *	    -EINVAL - accept syscall failed
  *	    -ENOMEM - cipher handle cannot be allocated
  */
-DSO_PUBLIC
 int kcapi_cipher_init(struct kcapi_handle **handle, const char *ciphername,
 		      uint32_t flags);
 
@@ -116,7 +101,6 @@ int kcapi_cipher_init(struct kcapi_handle **handle, const char *ciphername,
  *
  * @handle: [in] cipher handle to release
  */
-DSO_PUBLIC
 void kcapi_cipher_destroy(struct kcapi_handle *handle);
 
 /**
@@ -135,12 +119,11 @@ void kcapi_cipher_destroy(struct kcapi_handle *handle);
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int kcapi_cipher_setkey(struct kcapi_handle *handle,
 			const uint8_t *key, uint32_t keylen);
 
 /**
- * kcapi_cipher_encrypt() - encrypt data (one shot)
+ * kcapi_cipher_encrypt() - encrypt data (synchronous one shot)
  *
  * @handle: [in] cipher handle
  * @in: [in] plaintext data buffer
@@ -165,14 +148,41 @@ int kcapi_cipher_setkey(struct kcapi_handle *handle,
  * @return number of bytes encrypted upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
 			     const uint8_t *iv,
 			     uint8_t *out, uint32_t outlen, int access);
 
 /**
- * kcapi_cipher_decrypt() - decrypt data (one shot)
+ * kcapi_cipher_encrypt_aio() - encrypt data (asynchronous one shot)
+ *
+ * @handle: [in] cipher handle
+ * @iov: [in/out] head of scatter-gather list array holding the plaintext
+ *	which will be replaced with ciphertext
+ * @iovlen: [in] number of scatter-gather list entries
+ * @iv: [in] IV to be used for cipher operation
+ * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
+ *	heuristic for  fastest kernel access; KCAPI_ACCESS_VMSPLICE - use
+ *	vmsplice access; KCAPI_ACCESS_SENDMSG - sendmsg access)
+ *
+ * The individual scatter-gather list entries are processed with
+ * separate invocations of the the given cipher.
+ *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(sysconf(_SC_PAGESIZE)), If it is not aligned at the page
+ * boundary, the vmsplice call may not send all data to the kernel.
+ *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
+ * @return number of bytes encrypted upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_cipher_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+				 uint32_t iovlen, const uint8_t *iv,
+				 int access);
+
+/**
+ * kcapi_cipher_decrypt() - decrypt data (synchronous one shot)
  *
  * @handle: [in] cipher handle
  * @in: [in] ciphertext data buffer
@@ -197,12 +207,39 @@ int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * @return number of bytes decrypted upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
 			     const uint8_t *iv,
 			     uint8_t *out, uint32_t outlen, int access);
 
+/**
+ * kcapi_cipher_decrypt_aio() - decrypt data (asynchronous one shot)
+ *
+ * @handle: [in] cipher handle
+ * @iov: [in/out] head of scatter-gather list array holding the ciphertext
+ *	which will be replaced with plaintext
+ * @iovlen: [in] number of scatter-gather list entries
+ * @iv: [in] IV to be used for cipher operation
+ * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
+ *	heuristic for  fastest kernel access; KCAPI_ACCESS_VMSPLICE - use
+ *	vmsplice access; KCAPI_ACCESS_SENDMSG - sendmsg access)
+ *
+ * The individual scatter-gather list entries are processed with
+ * separate invocations of the the given cipher.
+ *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(sysconf(_SC_PAGESIZE)), If it is not aligned at the page
+ * boundary, the vmsplice call may not send all data to the kernel.
+ *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
+ * @return number of bytes decrypted upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_cipher_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+				 uint32_t iovlen, const uint8_t *iv,
+				 int access);
+
 /**
  * kcapi_cipher_stream_init_enc() - start an encryption operation (stream)
  *
@@ -232,7 +269,6 @@ int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 				     const uint8_t *iv,
 				     struct iovec *iov, uint32_t iovlen);
@@ -265,7 +301,6 @@ int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 				     const uint8_t *iv,
 				     struct iovec *iov, uint32_t iovlen);
@@ -302,7 +337,6 @@ int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
 				   struct iovec *iov, uint32_t iovlen);
 
@@ -328,7 +362,6 @@ int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
  * @return number of bytes obtained from the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
 			       struct iovec *iov, uint32_t iovlen);
 
@@ -340,7 +373,6 @@ int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
  * @return > 0 specifying the IV size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_cipher_ivsize(struct kcapi_handle *handle);
 
 /**
@@ -351,7 +383,6 @@ uint32_t kcapi_cipher_ivsize(struct kcapi_handle *handle);
  * @return > 0 specifying the block size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle);
 
 
@@ -379,7 +410,6 @@ uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle);
  *	    -EINVAL - accept syscall failed
  *	    -ENOMEM - cipher handle cannot be allocated
  */
-DSO_PUBLIC
 int kcapi_aead_init(struct kcapi_handle **handle, const char *ciphername,
 		    uint32_t flags);
 
@@ -388,7 +418,6 @@ int kcapi_aead_init(struct kcapi_handle **handle, const char *ciphername,
  *
  * @handle: [in] cipher handle to release
  */
-DSO_PUBLIC
 void kcapi_aead_destroy(struct kcapi_handle *handle);
 
 /**
@@ -407,7 +436,6 @@ void kcapi_aead_destroy(struct kcapi_handle *handle);
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int kcapi_aead_setkey(struct kcapi_handle *handle,
 		      const uint8_t *key, uint32_t keylen);
 
@@ -423,7 +451,6 @@ int kcapi_aead_setkey(struct kcapi_handle *handle,
  * @return 0 upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int kcapi_aead_settaglen(struct kcapi_handle *handle, uint32_t taglen);
 
 /**
@@ -438,11 +465,10 @@ int kcapi_aead_settaglen(struct kcapi_handle *handle, uint32_t taglen);
  * data size and may return an error during initialization if the
  * authentication size is not considered appropriate.
  */
-DSO_PUBLIC
 void kcapi_aead_setassoclen(struct kcapi_handle *handle, uint32_t assoclen);
 
 /**
- * kcapi_aead_encrypt() - encrypt AEAD data (one shot)
+ * kcapi_aead_encrypt() - synchronously encrypt AEAD data (one shot)
  *
  * @handle: [in] cipher handle
  * @in: [in] plaintext data buffer
@@ -479,13 +505,50 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, uint32_t assoclen);
  * @return number of bytes encrypted upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
 			   const uint8_t *iv,
 			   uint8_t *out, uint32_t outlen,
 			   int access);
 
+/**
+ * kcapi_aead_encrypt_aio() - asynchronously encrypt AEAD data (one shot)
+ *
+ * @handle: [in] cipher handle
+ * @iov: [in/out] array of scatter-gather list with input / output buffers
+ * @iovlen: [in] number of IOVECs in array
+ * @iv: [in] IV to be used for cipher operation
+ * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
+ *	heuristic for fastest kernel access; KCAPI_ACCESS_VMSPLICE - use
+ *	vmsplice access; KCAPI_ACCESS_SENDMSG - sendmsg access)
+ *
+ * The AEAD cipher operation requires the furnishing of the associated
+ * authentication data. In case such data is not required, it can be set to
+ * NULL and length value must be set to zero.
+ *
+ * Each IOVEC is processed with its individual AEAD cipher operation. The
+ * memory holding the input data will receive the processed data.
+ *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(sysconf(_SC_PAGESIZE)), If it is not aligned at the page
+ * boundary, the vmsplice call may not send all data to the kernel.
+ *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
+ * After invoking this function the caller should use
+ * kcapi_aead_getdata() to obtain the resulting ciphertext and authentication
+ * tag references.
+ *
+ * IMPORTANT NOTE: The kernel will only process
+ * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
+ * be handled by the kernel.
+ *
+ * @return number of bytes encrypted upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_aead_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+			       uint32_t iovlen, const uint8_t *iv, int access);
+
 /**
  * kcapi_aead_getdata() - get the resulting data from encryption
  *
@@ -511,7 +574,6 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
  * provides the consumer with the length of the tag and the length of the
  * ciphertext.
  */
-DSO_PUBLIC
 void kcapi_aead_getdata(struct kcapi_handle *handle,
 			uint8_t *encdata, uint32_t encdatalen,
 			uint8_t **aad, uint32_t *aadlen,
@@ -519,7 +581,7 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
 			uint8_t **tag, uint32_t *taglen);
 
 /**
- * kcapi_aead_decrypt() - decrypt AEAD data (one shot)
+ * kcapi_aead_decrypt() - synchronously decrypt AEAD data (one shot)
  *
  * @handle: [in] cipher handle
  * @in: [in] ciphertext data buffer
@@ -557,12 +619,50 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  * @return number of bytes decrypted upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
 			   const uint8_t *iv,
 			   uint8_t *out, uint32_t outlen, int access);
 
+/**
+ * kcapi_aead_decrypt_aio() - asynchronously decrypt AEAD data (one shot)
+ *
+ * @handle: [in] cipher handle
+ * @iov: [in/out] array of scatter-gather list with input / output buffers
+ * @iovlen: [in] number of IOVECs in array
+ * @iv: [in] IV to be used for cipher operation
+ * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
+ *	heuristic for fastest kernel access; KCAPI_ACCESS_VMSPLICE - use
+ *	vmsplice access; KCAPI_ACCESS_SENDMSG - sendmsg access)
+ *
+ * The AEAD cipher operation requires the furnishing of the associated
+ * authentication data. In case such data is not required, it can be set to
+ * NULL and length value must be set to zero.
+ *
+ * Each IOVEC is processed with its individual AEAD cipher operation. The
+ * memory holding the input data will receive the processed data.
+ *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(sysconf(_SC_PAGESIZE)), If it is not aligned at the page
+ * boundary, the vmsplice call may not send all data to the kernel.
+ *
+ * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
+ *
+ * To catch authentication errors (i.e. integrity violations) during the
+ * decryption operation, the errno of this call shall be checked for EBADMSG.
+ * If this function returns < 0 and errno is set to EBADMSG, an authentication
+ * error is detected.
+ *
+ * IMPORTANT NOTE: The kernel will only process
+ * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
+ * be handled by the kernel.
+ *
+ * @return number of bytes encrypted upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_aead_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+			       uint32_t iovlen, const uint8_t *iv, int access);
+
 /**
  * kcapi_aead_stream_init_enc() - start an encryption operation (stream)
  *
@@ -600,7 +700,6 @@ int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 				   const uint8_t *iv,
 				   struct iovec *iov, uint32_t iovlen);
@@ -641,7 +740,6 @@ int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 				   const uint8_t *iv,
 				   struct iovec *iov, uint32_t iovlen);
@@ -680,7 +778,6 @@ int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
 				 struct iovec *iov, uint32_t iovlen);
 
@@ -702,7 +799,6 @@ int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
 				      struct iovec *iov, uint32_t iovlen);
 
@@ -725,7 +821,6 @@ int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
  * @return number of bytes obtained from the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			     struct iovec *iov, uint32_t iovlen);
 
@@ -737,7 +832,6 @@ int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
  * @return > 0 specifying the IV size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_aead_ivsize(struct kcapi_handle *handle);
 
 /**
@@ -748,7 +842,6 @@ uint32_t kcapi_aead_ivsize(struct kcapi_handle *handle);
  * @return > 0 specifying the block size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_aead_blocksize(struct kcapi_handle *handle);
 
 /**
@@ -763,7 +856,6 @@ uint32_t kcapi_aead_blocksize(struct kcapi_handle *handle);
  * @return > 0 specifying the block size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_aead_authsize(struct kcapi_handle *handle);
 
 /**
@@ -776,7 +868,6 @@ uint32_t kcapi_aead_authsize(struct kcapi_handle *handle);
  *
  * @return minimum size of output data length in bytes
  */
-DSO_PUBLIC
 uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
 			    uint32_t inlen, uint32_t assoclen, uint32_t taglen);
 
@@ -796,7 +887,6 @@ uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
  * @return 0 upon success;
  *	    < 0 upon failure
  */
-DSO_PUBLIC
 int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
 			       uint8_t **iv, uint32_t *ivlen);
 
@@ -820,7 +910,6 @@ int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
  *	    -EINVAL - accept syscall failed;
  *	    -ENOMEM - cipher handle cannot be allocated
  */
-DSO_PUBLIC
 int kcapi_md_init(struct kcapi_handle **handle, const char *ciphername,
 		  uint32_t flags);
 
@@ -829,7 +918,6 @@ int kcapi_md_init(struct kcapi_handle **handle, const char *ciphername,
  *
  * @handle: [in] cipher handle to release
  */
-DSO_PUBLIC
 void kcapi_md_destroy(struct kcapi_handle *handle);
 
 /**
@@ -848,7 +936,6 @@ void kcapi_md_destroy(struct kcapi_handle *handle);
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int kcapi_md_setkey(struct kcapi_handle *handle,
 		    const uint8_t *key, uint32_t keylen);
 
@@ -862,7 +949,6 @@ int kcapi_md_setkey(struct kcapi_handle *handle,
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int32_t kcapi_md_update(struct kcapi_handle *handle,
 			const uint8_t *buffer, uint32_t len);
 
@@ -878,7 +964,6 @@ int32_t kcapi_md_update(struct kcapi_handle *handle,
  * 	    -ENOMEM - buffer is too small for the complete message digest,
  * 	    the buffer is filled with the truncated message digest
  */
-DSO_PUBLIC
 int32_t kcapi_md_final(struct kcapi_handle *handle,
 		       uint8_t *buffer, uint32_t len);
 
@@ -903,7 +988,6 @@ int32_t kcapi_md_final(struct kcapi_handle *handle,
  * 	    -ENOMEM - buffer is too small for the complete message digest,
  * 	    the buffer is filled with the truncated message digest
  */
-DSO_PUBLIC
 int32_t kcapi_md_digest(struct kcapi_handle *handle,
 		       const uint8_t *in, uint32_t inlen,
 		       uint8_t *out, uint32_t outlen);
@@ -920,7 +1004,6 @@ int32_t kcapi_md_digest(struct kcapi_handle *handle,
  * @return > 0 specifying the block size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_md_digestsize(struct kcapi_handle *handle);
 
 /**
@@ -931,7 +1014,6 @@ uint32_t kcapi_md_digestsize(struct kcapi_handle *handle);
  * @return > 0 specifying the block size;
  *	    0 on error
  */
-DSO_PUBLIC
 uint32_t kcapi_md_blocksize(struct kcapi_handle *handle);
 
 
@@ -955,7 +1037,6 @@ uint32_t kcapi_md_blocksize(struct kcapi_handle *handle);
  *	    -EINVAL - accept syscall failed
  *	    -ENOMEM - cipher handle cannot be allocated
  */
-DSO_PUBLIC
 int kcapi_rng_init(struct kcapi_handle **handle, const char *ciphername,
 		   uint32_t flags);
 
@@ -964,7 +1045,6 @@ int kcapi_rng_init(struct kcapi_handle **handle, const char *ciphername,
  *
  * @handle: [in] cipher handle to release
  */
-DSO_PUBLIC
 void kcapi_rng_destroy(struct kcapi_handle *handle);
 
 /**
@@ -985,7 +1065,6 @@ void kcapi_rng_destroy(struct kcapi_handle *handle);
  * @return 0 upon success;
  * 	    < 0 upon error
  */
-DSO_PUBLIC
 int kcapi_rng_seed(struct kcapi_handle *handle, uint8_t *seed,
 		   uint32_t seedlen);
 
@@ -999,7 +1078,6 @@ int kcapi_rng_seed(struct kcapi_handle *handle, uint8_t *seed,
  * @return size of random number generated upon success;
  *	    -EIO - data cannot be obtained
  */
-DSO_PUBLIC
 int32_t kcapi_rng_generate(struct kcapi_handle *handle,
 			   uint8_t *buffer, uint32_t len);
 
@@ -1025,7 +1103,6 @@ enum kcapi_verbosity {
  *	LOG_VERBOSE: log verbose messages, warnings and error messages
  *	LOG_DEBUG: log all details of library operation
  */
-DSO_PUBLIC
 void kcapi_set_verbosity(enum kcapi_verbosity level);
 
 /**
@@ -1034,7 +1111,6 @@ void kcapi_set_verbosity(enum kcapi_verbosity level);
  * @buf: [out] buffer to place version string into
  * @buflen: [in] length of buffer
  */
-DSO_PUBLIC
 void kcapi_versionstring(char *buf, uint32_t buflen);
 
 /**
@@ -1050,7 +1126,6 @@ void kcapi_versionstring(char *buf, uint32_t buflen);
  *
  * @return Version number of kcapi library
  */
-DSO_PUBLIC
 uint32_t kcapi_version(void);
 
 /**
@@ -1072,7 +1147,6 @@ uint32_t kcapi_version(void);
  * @return 0 for success;
  *	    < 0 for any errors
  */
-DSO_PUBLIC
 int kcapi_pad_iv(struct kcapi_handle *handle,
 		 const uint8_t *iv, uint32_t ivlen,
 		 uint8_t **newiv, uint32_t *newivlen);
@@ -1087,13 +1161,12 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
  *
  * The parameters, he logic and the return code is identical to memset(3).
  */
-DSO_PUBLIC
 void kcapi_memset_secure(void *s, int c, uint32_t n);
 
 /**
  * DOC: Asymmetric Cipher API
  *
- * API function calls used to invoke symmetric ciphers.
+ * API function calls used to invoke asymmetric ciphers.
  */
 
 /**
@@ -1112,7 +1185,6 @@ void kcapi_memset_secure(void *s, int c, uint32_t n);
  *	    -EINVAL - accept syscall failed
  *	    -ENOMEM - cipher handle cannot be allocated
  */
-DSO_PUBLIC
 int kcapi_akcipher_init(struct kcapi_handle **handle, const char *ciphername,
 			uint32_t flags);
 
@@ -1121,7 +1193,6 @@ int kcapi_akcipher_init(struct kcapi_handle **handle, const char *ciphername,
  *
  * @handle: [in] cipher handle to release
  */
-DSO_PUBLIC
 void kcapi_akcipher_destroy(struct kcapi_handle *handle);
 
 /**
@@ -1153,7 +1224,6 @@ void kcapi_akcipher_destroy(struct kcapi_handle *handle);
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int kcapi_akcipher_setkey(struct kcapi_handle *handle,
 			  const uint8_t *key, uint32_t keylen);
 
@@ -1179,7 +1249,6 @@ int kcapi_akcipher_setkey(struct kcapi_handle *handle,
  * @return 0 upon success;
  *	    < 0 in case of error
  */
-DSO_PUBLIC
 int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
 			     const uint8_t *key, uint32_t keylen);
 
@@ -1209,7 +1278,6 @@ int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
  * @return number of bytes returned by the encryption operation upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
 			       uint8_t *out, uint32_t outlen, int access);
@@ -1240,7 +1308,6 @@ int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
  * @return number of bytes returned by the decryption operation upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
 			       uint8_t *out, uint32_t outlen, int access);
@@ -1271,7 +1338,6 @@ int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
  * @return number of bytes returned by the signature gen operation upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
 			    const uint8_t *in, uint32_t inlen,
 			    uint8_t *out, uint32_t outlen, int access);
@@ -1306,7 +1372,6 @@ int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
  * @return number of bytes returned by the signature ver operation upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
 			      const uint8_t *in, uint32_t inlen,
 			      uint8_t *out, uint32_t outlen, int access);
@@ -1337,7 +1402,6 @@ int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
 
@@ -1366,7 +1430,6 @@ int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
 
@@ -1396,7 +1459,6 @@ int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
 
@@ -1427,7 +1489,6 @@ int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
 
@@ -1454,7 +1515,6 @@ int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
 				     struct iovec *iov, uint32_t iovlen);
 
@@ -1479,7 +1539,6 @@ int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
  * @return number of bytes sent to the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
 					  struct iovec *iov, uint32_t iovlen);
 
@@ -1506,8 +1565,152 @@ int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
  * @return number of bytes obtained from the kernel upon success;
  *	    < 0 in case of error with errno set
  */
-DSO_PUBLIC
 int32_t kcapi_akcipher_stream_op(struct kcapi_handle *handle,
 			         struct iovec *iov, uint32_t iovlen);
 
+/**
+ * DOC: Key Derivation Functions
+ *
+ * API function calls used to invoke a KDF.
+ * The KDF functions are based on a message digest or keyed message digest
+ * function. The caller must have the handle allocated with kcapi_md_init.
+ * If the caller wishes to use a keyed message digest, the caller must invoke
+ * kcapi_md_setkey before those functions.
+ */
+
+/**
+ * kcapi_kdf_dpi() - Double Pipeline Mode Key Derivation Function
+ *
+ * @handle: [in] cipher handle allocated by caller. This cipher handle
+ *	must be allocated with kcapi_md_init(). If the caller is interested in
+ *	a KDF using a keyed message digest, the caller should also call
+ *	kcapi_md_setkey() before invoking this function.
+ * @src: [in] Input data that should be transformed into a key (see below).
+ * @slen: [in] Length of the src input data.
+ * @dst: [out] Buffer to store the generated key in,
+ * @dlen: [in] Length of the dst buffer. This value defines the number of bytes
+ *	generated by the KDF.
+ *
+ * This function is an implementation of the KDF in double pipeline iteration
+ * mode according with counter to SP800-108 section 5.3.
+ *
+ * The caller must provide Label || 0x00 || Context in src. This src pointer
+ * may also be NULL if the caller wishes not to provide anything.
+ *
+ * @return 0 upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_kdf_dpi(struct kcapi_handle *handle,
+		      const uint8_t *src, uint32_t slen,
+		      uint8_t *dst, uint32_t dlen);
+
+/**
+ * kcapi_kdf_fb() - Feedback Mode Key Derivation Function
+ *
+ * @handle: [in] cipher handle allocated by caller. This cipher handle
+ *	must be allocated with kcapi_md_init(). If the caller is interested in
+ *	a KDF using a keyed message digest, the caller should also call
+ *	kcapi_md_setkey() before invoking this function.
+ * @src: [in] Input data that should be transformed into a key (see below).
+ * @slen: [in] Length of the src input data.
+ * @dst: [out] Buffer to store the generated key in,
+ * @dlen: [in] Length of the dst buffer. This value defines the number of bytes
+ *	generated by the KDF.
+ *
+ * This function is an implementation of the KDF in feedback mode with a
+ * non-NULL IV and with counter according to SP800-108 section 5.2. The IV is
+ * supplied with src and must be equal to the digestsize of the used cipher.
+ *
+ * In addition, the caller must provide Label || 0x00 || Context in src. This
+ * src pointer must not be NULL as the IV is required. The ultimate format of
+ * the src pointer is IV || Label || 0x00 || Context where the length of the
+ * IV is equal to the block size (i.e. the digest size of the underlying
+ * hash) of the PRF.
+ *
+ * @return 0 upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_kdf_fb(struct kcapi_handle *handle,
+		     const uint8_t *src, uint32_t slen,
+		     uint8_t *dst, uint32_t dlen);
+
+/**
+ * kcapi_kdf_ctr() - Counter Mode Key Derivation Function
+ *
+ * @handle: [in] cipher handle allocated by caller. This cipher handle
+ *	must be allocated with kcapi_md_init(). If the caller is interested in
+ *	a KDF using a keyed message digest, the caller should also call
+ *	kcapi_md_setkey() before invoking this function.
+ * @src: [in] Input data that should be transformed into a key (see below).
+ * @slen: [in] Length of the src input data.
+ * @dst: [out] Buffer to store the generated key in,
+ * @dlen: [in] Length of the dst buffer. This value defines the number of bytes
+ *	generated by the KDF.
+ *
+ * This function is an implementation of the KDF in counter mode according to
+ * SP800-108 section 5.1 as well as SP800-56A section 5.8.1 (Single-step KDF).
+ *
+ * SP800-108:
+ * The caller must provide Label || 0x00 || Context in src. This src pointer
+ * may also be NULL if the caller wishes not to provide anything.
+ *
+ * SP800-56A:
+ * If a keyed MAC is used, the key shall NOT be the shared secret from the DH
+ * operation, but an independently generated key. The src pointer is defined
+ * as Z || other info where Z is the shared secret from DH and other info is an
+ * arbitrary string (see SP800-56A section 5.8.1.2).
+ *
+ * @return 0 upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_kdf_ctr(struct kcapi_handle *handle,
+		      const uint8_t *src, uint32_t slen,
+		      uint8_t *dst, uint32_t dlen);
+
+/**
+ * kcapi_pbkdf() - Password-based Key Derivation Function
+ *
+ * @hashname: [in] kernel crypto API name of a keyed hash (e.g. hmac(sha1))
+ * @pw: [in] Password a key shall be derived from
+ * @pwlen: [in] Length of password string
+ * @salt: [in] Salt as defined in SP800-132
+ * @saltlen: [in] Length of salt buffer
+ * @count: [in] Numbers of iterations to be performed for the PBKDF
+ * @key: [out] Buffer to store the generated key in
+ * @keylen: [in] Size of the key to be generated (i.e. length of the key buffer)
+ *
+ * This function is an implementation of the PBKDF as defined in SP800-132.
+ *
+ * @return 0 upon success;
+ *	    < 0 in case of error with errno set
+ */
+int32_t kcapi_pbkdf(const char *hashname,
+		    const uint8_t *pw, uint32_t pwlen,
+		    const uint8_t *salt, uint32_t saltlen,
+		    uint32_t count,
+		    uint8_t *key, uint32_t keylen);
+
+/**
+ * kcapi_pbkdf_iteration_count() - Calculate numbers of iterations for a PBKDF
+ *
+ * @hashname: [in] kernel crypto API name of a keyed hash (e.g. hmac(sha1))
+ * @timeshresh: [in] Time duration in nanoseconds that the PBKDF operation
+ *	shall at least require. If that value is 0, a default of (1<<27)
+ *	nanoseconds is used.
+ *
+ * The function measures the time the PBKDF operation takes for different
+ * round counts for the given keyed message digest type.
+ *
+ * The result should be taken as the iteration count for a PBKDF operation.
+ *
+ * If an error occurs with the PBKDF calculation, a value of 1<<18 is returned.
+ *
+ * @return number of iterations a PBKDF should take on this computer.
+ */
+uint32_t kcapi_pbkdf_iteration_count(const char *hashname, uint64_t timeshresh);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* _KCAPI_H */
